header {* The Monad class *}

theory Monad
imports MonadBase
begin

text {* 
  Kleisli-application. We need one variant for each arity, unfortunately.
  *}
constdefs 
  kapp :: "('a=> 'b M)=> 'a M=> 'b M"        
 "kapp f x ==  {x<- x; f x}"
  kapp2 :: "('a=> 'b => 'c M)=> 'a M=> 'b M=> 'c M"        
 "kapp2 f x y == {x<- x; y<- y; f x y}"
  kapp3 :: "('a=> 'b => 'c => 'd M)=> 'a M=> 'b M=> 'c M=> 'd M"
 "kapp3 f x y z == {x<- x; y<- y; z<- z; f x y z}"

text {* 
  This sets up a uniform Kleisli application, i.e. 
  @term{f $$ x $$ y $$ z} for ternary operations. (Only up to arity 
  3 at the moment, I am afraid.) 
  *} 

syntax
 "_kapp"  :: "'x => 'y M=> 'z M"  (infixl "$$" 5)

translations
  "_kapp(_kapp(_kapp f x) y) z" == "kapp3 f x y z" 
  "_kapp(_kapp f x) y" == "kapp2 f x y" 
  "_kapp f x"  ==  "kapp f x" 


section{* Lifting *}

text{*
  Corresponds to the free functor from the base into the Kleisli category.
  We need one lift for each arity, due to currying.
  These are designed to work well with Kleisli application, i.e. you can write
  things like @term{lift2 f $$ a $$ b}.
  *}

constdefs 
  lift :: "('a => 'b) => 'a => 'b M" 
 "lift f == eta o f"
  lift2 :: "('a => 'b=> 'c) => 'a => 'b => 'c M"
 "lift2 f == %a. lift (f a)"
  lift3 :: "('a => 'b=> 'c=> 'd) => 'a => 'b=> 'c=> 'd M"
 "lift3 f a b == lift (f a b)" 

lemma lift_eta:  "(lift f $$ (eta x)) = eta (f x)"
 by (unfold lift_def kapp_def, simp)

lemma lift_eta2: "(lift f $$ ({p; q})) = ({p; lift f $$ q})"
 by (unfold lift_def kapp_def, simp)

lemma lift2_eta: "(lift2 f $$ (eta x) $$ (eta y)) = (eta (f x y))"
 by (simp add: lift2_def kapp2_def, 
     simp add: lift_def kapp_def)

lemma lift3_eta: "(lift3 f $$ (eta x) $$ (eta y)$$ (eta z)) = (eta (f x y z))"
 by (simp add: lift3_def kapp3_def, 
     simp add: lift_def kapp_def)


end