
signature DATATYPE_RECORD =
sig

type InstStruct

datatype Result = Result of {ids : string list,
                             data : InstStruct,
                             intro_consts : string list,
                             class_deps : string list,
                             tycon_deps : string list}

val try_datatype : theory -> string -> Result option
val try_record   : theory -> string -> Result option

val translate_datatype_record : 
((string * mixfix option) Symtab.table -> Context.theory -> 
 (TypeS.TS Symtab.table -> Term.typ -> Term.typ) ->
  TypeS.TS Symtab.table -> Symtab.key Symtab.table ->
  Derive_Thm.AxmImg Symtab.table -> InstStruct ->
 (theory * TypeS.TS Symtab.table * Symtab.key Symtab.table * Derive_Thm.AxmImg Symtab.table)) option

end


structure Datatype_Record =
struct



datatype InstStruct = Datatype of {num : int,
                                   types : (int * (string list * string * mixfix)) list,
                                   constrs : (int * ((string * Term.typ * mixfix) list)) list}
                    | Record of {qname : string,
                                 recname : string,
                                 args : string list,
                                 arity : int,
                                 parent : (Term.typ list * string) option,
                                 fields : (string * Term.typ * mixfix) list}

datatype Result = Result of {ids : string list,
                             data : InstStruct,
                             intro_consts : string list,
                             class_deps : string list,
                             tycon_deps : string list}



fun try_datatype thy tid =
(case (DatatypePackage.get_datatype thy tid) of
  SOME {index, descr, induction, inject, distinct, ...} => 
   (let fun get_deps1 (DatatypeAux.DtTFree x) = []
          | get_deps1 (DatatypeAux.DtType (f, xs)) = Aux.ins_string f (Aux.union_strs (map get_deps1 xs))
          | get_deps1 (DatatypeAux.DtRec i) = []
        fun get_deps2 [] = []
          | get_deps2 ((c, args)::cs) = Aux.union_str(Aux.union_strs (map get_deps1 args), get_deps2 cs) 

        fun get_arity [] = []
          | get_arity (x::xs) = (case x of
                                  DatatypeAux.DtTFree v => v::(get_arity xs)
                                | _ => error ("Unexpected constructor in get_arity"))
        fun prep_constrs [] = []
          | prep_constrs ((c, ty)::cs) = 
                                     (case (Syn_mng.get_mixfix c) of
                                           NONE => (c, Aux.unvarifyT ty, NoSyn)::(prep_constrs cs)
                                         | SOME m => (c, Aux.unvarifyT ty, m)::(prep_constrs cs))


        fun get_data ids pdeps types constrs [] = (ids, pdeps, types, constrs)
          | get_data ids pdeps types constrs ((ind, (nm, args, cs))::dsc) = 
            let val mxfx = case (Syn_mng.get_type_mixfix nm) of
                             NONE => NoSyn
                           | SOME m => m
                val arity = get_arity args
                val deps = get_deps2 cs
                val pcs = (case (DatatypePackage.get_datatype_constrs thy nm) of
                              NONE => error ("Unexpected error: constrs_of " ^ nm)
                            | SOME cs_pr => prep_constrs cs_pr)
            in
              get_data (nm::ids) (Aux.union_str(deps, pdeps)) ((ind, (arity, nm, mxfx))::types) 
                       ((ind, pcs)::constrs) dsc
            end
        val (ids, pdeps, types, constrs) = get_data [] [] [] [] descr
        val deps = Aux.diff_str(pdeps, ids)
        val num = (length descr)  

        val data = Datatype {num = num, types = types, constrs = constrs}

        fun get_intros [] = []
          | get_intros (x::xs) = 
           (let val (q, base_id) = (Aux.qualifier x, Aux.base x)
            in
             (q ^ ".Abs_" ^ base_id)::(q ^ ".Rep_" ^ base_id)::(get_intros xs)
            end)

        val ext = Result {ids = ids, data = data, intro_consts = [], (* not required yet: get_intros ids, *)
                          class_deps = [], tycon_deps = deps}
   in
      SOME ext
   end)
| NONE => NONE)

fun try_record thy tid =
      (case (try (unsuffix RecordPackage.ext_typeN) tid) of
         SOME x =>
                                    (* qualifier <thy>.<recname>.<recname>_ext_type = <thy>.<recname>
                                       qualifier <thy>.<recname>_scheme = <thy> *) 
          (case (RecordPackage.get_extension thy (Aux.qualifier tid)) of
              SOME info =>
                  (let val qname = Aux.qualifier tid
                       val recnm = Aux.base qname
                       val (a, i) = case (Aux.type_arity (thy, tid)) of
                                          NONE => error ("Unexpected exception with type_arity in try_record")
                                        | SOME (a, i) => (a, i) 


                       val (tc, args) = info
                       val src_rec_type = Type (tc, args)
                       val recT_fields = fst(RecordPackage.get_recT_fields thy src_rec_type)
                       
                       fun extract_args [] = error "Unexpected empty arguments for extract_args"
                         | extract_args [u] = [] 
                         | extract_args (x::xs) = (case x of
                                                    TFree (v, s) => (v::(extract_args xs))
                                                   | _ => error ("Unexpected constructor of a \
                                                                  \ record-type argument in extract_args"))
                       val arity = extract_args args
                       val parent = RecordPackage.get_parent thy qname 

                       fun un_str' (a : string list, b) = case b of 
                                                            NONE => a
                                                          | SOME x => a union x
                       fun extract_deps (cs, tcs) [] = (cs, tcs)
                         | extract_deps (cs, tcs) ((sel, tp)::flds) = 
                            (let val cs' = Aux.all_classes_of_type tp
                                 val tcs' = Aux.all_tycons_of_type tp
                             in
                              extract_deps (Aux.union_str(cs, cs'), Aux.union_str(tcs, tcs')) flds
                             end)
                       val (cs, tcs) = extract_deps ([], []) recT_fields
                       val (cs', tcs') = 
                    (un_str' (cs, Option.map (Aux.all_classes_of_type o (fn (tp, f) => Type (f, tp))) parent),
                     un_str' (tcs, Option.map (Aux.all_tycons_of_type o (fn (tp, f) => Type (f, tp))) parent))

                       fun prep_fields [] = []
                         | prep_fields ((sel, tp)::flds) =
                           (case (Syn_mng.get_mixfix sel) of
                              NONE => (sel, tp, NoSyn)::(prep_fields flds)
                            | SOME mx => (sel, tp, mx)::(prep_fields flds))
                       val fields = prep_fields recT_fields

                       val recdata = Record {qname = qname, recname = recnm,
                                             args = arity, arity = i, parent = parent,
                                             fields = fields}
                       val extdata = Result{ids = [tid], data = recdata, intro_consts = [],
                                            class_deps = cs', tycon_deps = tcs'} 
                   in
                    SOME extdata
                   end)
            | NONE => error ("Unexpected error: no extension for " ^ x))
       | NONE => NONE)



fun translate_record renames trg map_type typeMap opMap axmMap 
         (Record{qname, recname, args, arity, parent, fields}) =
 let val trg_id = Aux.name_of_thy trg
     val recname' = (case (Symtab.lookup renames qname) of
                       NONE => Aux.base recname
                     | SOME (r, _) => r)
     val qname' = trg_id ^ "." ^ recname'

     fun map_parent (ts, p) = (let val ts' = map (map_type typeMap) ts
                                   val p' = (case (Symtab.lookup renames p) of
                                                 NONE => (trg_id ^ "." ^ (Aux.base p))
                                               | SOME (x, _) => (trg_id ^ "." ^ x))
                                  in
                                   (ts', p')
                                  end)

     val parent' = Option.map map_parent parent

     fun map_fields [] = []
       | map_fields ((sel, ty, mxfx')::xs) =
         let val (nsel, mxfx) = (case (Symtab.lookup renames sel) of
                                  NONE => (case (Syn_mng.get_mixfix sel) of
                                             NONE => (Aux.base sel, mxfx')
                                           | SOME m => (Aux.base sel, m))
                                | SOME (nm, mx) => (case mx of
                                                      NONE => (nm, mxfx')
                                                    | SOME m => (nm, m)))
             val _ = (case mxfx of
                       NoSyn => ()
                     | _ => Syn_mng.add_mixfix (qname' ^ "." ^ nsel, mxfx))
             val ty' = map_type typeMap ty
         in
           (nsel, ty', mxfx)::(map_fields xs)
         end
      val fields' = map_fields fields

      val tid = qname ^ "." ^ recname ^ RecordPackage.ext_typeN        
      val tid' = trg_id ^ "." ^ recname' ^ "." ^ recname' ^ RecordPackage.ext_typeN 
 in
  case (Symtab.lookup typeMap tid) of
   NONE =>
 (let val typeMap' = Symtab.update_new (tid, TypeS.typeS_of_TC arity tid') typeMap
                     handle DUP => error ("Unexpected DUP in translate_record, typeMap: " ^ tid)
      val opMap' = Symtab.update (qname ^ ".Abs_" ^ recname ^ RecordPackage.ext_typeN,
                                      qname' ^ ".Abs_" ^ recname' ^ RecordPackage.ext_typeN)
                   (Symtab.update (qname ^ ".Rep_" ^ recname ^ RecordPackage.ext_typeN,
                                       qname' ^ ".Rep_" ^ recname' ^ RecordPackage.ext_typeN) opMap)

                    handle DUP => error ("Unexpected DUP in translate_record, opMap")

      val axmMap' =  Symtab.update 
                      (qname ^ ".type_definition_" ^ recname ^ RecordPackage.ext_typeN,
                       Derive_Thm.Axm (qname' ^ ".type_definition_" ^ recname' ^ RecordPackage.ext_typeN)) axmMap
                    handle DUP => error ("Unexpected DUP in translate_record, axmMap #1")
      val axmMap'' =  Symtab.update 
                      (qname ^ ".arity_type_" ^ recname ^ RecordPackage.ext_typeN,
                       Derive_Thm.Axm (qname' ^ ".arity_type_" ^ recname' ^ RecordPackage.ext_typeN)) axmMap'
                    handle DUP => error ("Unexpected DUP in translate_record, axmMap #2")

      (* Adding record *)  
      val _ = Aux.info ("... adding record " ^ qname ^ 
                          " from the domain as " ^ recname' ^ " to the codomain ...")

      val trg' = trg 
                 |> RecordPackage.add_record_i false (args, recname') parent' fields'  

 in
  (trg', typeMap', opMap', axmMap'')
 end)
 
 | _ => (trg, typeMap, opMap, axmMap)
 end



fun translate_datatype renames trg map_type typeMap opMap axmMap (Datatype{num, types, constrs}) =
 let val trg_id = Aux.name_of_thy trg

     fun extract_args typemap (Type ("fun", [l, t])) = (map_type typemap l)::(extract_args typemap t)
       | extract_args typemap _ = []

     fun prep_constrs tid typemap [] = []
       | prep_constrs tid typemap ((c, ty, mxfx)::cs) =
         (let val (c', mxfx') = (case (Symtab.lookup renames c) of
                                  SOME (r, m) => (case m of
                                                   NONE => (r, mxfx)
                                                 | SOME mx => (r, mx))
                                | NONE => (Aux.base c, mxfx))
              val args = extract_args typemap ty
              val _ = (case mxfx' of
                        NoSyn => ()
                      | _ => Syn_mng.add_mixfix (tid ^ "." ^ c', mxfx'))
          in
           (Binding.name c', args, mxfx')::(prep_constrs tid typemap cs)
          end) 

     fun prep_types typemap res [] = (typemap, res)
       | prep_types typemap res ((ind, (arity, id, mxfx))::tys) =
         (let val (tid, mxfx') = 
                           (case (Symtab.lookup renames id) of
                             SOME (r, m) => (case m of
                                              NONE => (r, mxfx)
                                            | SOME mx => (r, mx))
                           | NONE => (Aux.base id, mxfx))
              val tid' = trg_id ^ "." ^ tid
              val _ = (case mxfx' of
                         NoSyn => ()
                       | _ => Syn_mng.add_type_mixfix (tid', mxfx'))

              val typemap' = Symtab.update_new (id, TypeS.typeS_of_TC (length arity) tid') typemap
          in
           prep_types typemap' ((ind, ((arity, tid, mxfx'), id))::res) tys
          end)

     val (typeMap', types') = prep_types typeMap [] types

     fun prep opmap axmmap ids tids declrs [] = (opmap, axmmap, ids, tids, declrs)
       | prep opmap axmmap ids tids declrs (ind::is) = 
       (case (Aux.assoc_int (types', ind)) of
         NONE => error ("Unexpected index in translate_datatype #1")
       | SOME ((arity, tid, mxfx), id) => 
         (case (Aux.assoc_int (constrs, ind)) of
           NONE => error ("Unexpected index in translate_datatype #2")
         | SOME cs => (let val (q, base_id) = (Aux.qualifier id, Aux.base id)
                           val opmap' = Symtab.update (q ^ ".Abs_" ^ base_id,
                                                           trg_id ^ ".Abs_" ^ tid)
                                       (Symtab.update (q ^ ".Rep_" ^ base_id,
                                                           trg_id ^ ".Rep_" ^ tid) opmap)
                           
                           val axmmap' = Symtab.update (q ^ ".type_definition_" ^ base_id,
                                                   Derive_Thm.Axm (trg_id ^ ".type_definition_" ^ tid)) axmmap 
                           val axmmap'' = Symtab.update (q ^ ".arity_type_" ^ base_id,
                                                   Derive_Thm.Axm (trg_id ^ ".arity_type_" ^ tid)) axmmap' 
                           val tid' = trg_id ^ "." ^ tid
                           val cs' = prep_constrs tid' typeMap' cs
                           val decl = (arity, Binding.name tid, mxfx, cs')
                       in
                        prep opmap' axmmap'' (ids @ [id]) (tids @ [tid]) (declrs @ [decl]) is
                       end)))

     val inds = 0 upto (num - 1)      
     val (opMap', axmMap', ids, tids, declrs) = prep opMap axmMap [] [] [] inds 

 (* Adding datatype(s) *)  
     val _ = Aux.info ("... adding datatype(s) " ^ (commas ids) ^ 
                          " from the domain as " ^ (commas tids) ^ " to the codomain ...")
     val (_, trg') = trg 
                     |> DatatypePackage.add_datatype false false tids declrs
 in
  (trg', typeMap', opMap', axmMap')
 end




fun translate_struct renames trg map_type typeMap opMap axmMap data =
 (case data of
    Datatype d => 
      (let fun check_types [] = true
             | check_types ((ind, (arity, id, mxfx))::tys) =
               (case (Symtab.lookup typeMap id) of
                 NONE => check_types tys
               | _ => false)
       in
        if (check_types (#types d))
        then translate_datatype renames trg map_type typeMap opMap axmMap (Datatype d)
        else (trg, typeMap, opMap, axmMap)
       end)
  | Record r => translate_record renames trg map_type typeMap opMap axmMap (Record r))
          

val translate_datatype_record = SOME translate_struct





end