(*
 * Extension of some Isabelle/Isar-parsers in order to get possibly defined concrete syntax
 * $Id: parser_ext.sml 1166 2009-04-23 10:22:29Z maxim $
 *)

structure Parser_Ext =
struct

structure P = OuterParse and K = OuterKeyword;

fun save_mixfix0 thynm (s, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s, mxfx))
  in
   (s, mxfx)
  end 

fun save_mixfix thy (bind, s2, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (Sign.full_name thy bind, mxfx))
  in
   (bind, s2, mxfx)
  end 

fun save_mixfix' thy (s1, s2, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix ((Aux.name_of_thy thy)^"."^s1, mxfx))
  in
   (s1, s2, mxfx)
  end 

fun save_mixfix2 thynm ((s1, s2, mxfx), s3) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => Syn_mng.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   ((s1, s2, mxfx), s3)
  end 

fun save_mixfix3 thy (l1, l2) =
 let fun f [] = ()
       | f ((opt, (x, s))::l) = (case opt of
                                 NONE => (f l)
                               | SOME (bind, s2, mxfx) =>
                                  (let val _ = (case mxfx of 
                                                 NoSyn => ()
                                                | _ => Syn_mng.add_mixfix (Sign.full_name thy bind, mxfx))
                                   in
                                    f l
                                   end))
     val _ = f l2
 in
  (l1, l2)
 end


(* *** Basic *** *)

val mode_spec =
  (P.$$$ "output" >> K ("", false)) || P.name -- Scan.optional (P.$$$ "output" >> K false) true;

val opt_mode =
  Scan.optional (P.$$$ "(" |-- P.!!! (mode_spec --| P.$$$ ")")) Syntax.mode_default;




(* *** Types *** *)

fun add_typedecls' (bind, args, mx) thy =
  let val _ = (case mx of
                 NoSyn => ()
               | _ => Syn_mng.add_type_mixfix (Sign.full_name thy bind, mx))
  in
   ObjectLogic.typedecl (bind, args, mx) thy
  end   


val _ =
  OuterSyntax.command "typedecl" "type declaration" K.thy_decl
    (P.type_args -- P.binding -- P.opt_infix >> (fn ((args, a), mx) =>
      Toplevel.theory (add_typedecls' (a, args, mx) #> snd)));


fun add_tyabbrs' abbrs thy =
  let val thynm = Aux.name_of_thy thy
      val _ = map (fn (bind, args, T, mx) => 
                    (case mx of
                        NoSyn => ()
                       | _ => Syn_mng.add_type_mixfix (Sign.full_name thy bind, mx))) abbrs
  in
   Sign.add_tyabbrs abbrs thy
  end


val _ =
  OuterSyntax.command "types" "declare type abbreviations" K.thy_decl
    (Scan.repeat1
      (P.type_args -- P.binding -- (P.$$$ "=" |-- P.!!! (P.typ -- P.opt_infix')))
      >> (Toplevel.theory o add_tyabbrs' o
        map (fn ((args, a), (T, mx)) => (a, args, T, mx))));


(* *** Consts *** *)

fun add_consts' cl thy = Sign.add_consts (map (save_mixfix thy) cl) thy

val _ =
  OuterSyntax.command "consts" "declare constants" K.thy_decl
    (Scan.repeat1 P.const_binding >> (Toplevel.theory o add_consts'));


(* *** Constdefs *** *)

fun add_constdefs' cl thy = Constdefs.add_constdefs (save_mixfix3 thy cl) thy

(* old constdefs *)

val old_constdecl =
  P.binding --| P.where_ >> (fn x => (x, NONE, NoSyn)) ||
  P.binding -- (P.$$$ "::" |-- P.!!! P.typ >> SOME) -- P.opt_mixfix'
    --| Scan.option P.where_ >> P.triple1 ||
  P.binding -- (P.mixfix >> pair NONE) --| Scan.option P.where_ >> P.triple2;

val old_constdef = Scan.option old_constdecl -- (SpecParse.opt_thm_name ":" -- P.prop);

val structs =
  Scan.optional ((P.$$$ "(" -- P.$$$ "structure") |-- P.!!! (P.simple_fixes --| P.$$$ ")")) [];

val _ =
  OuterSyntax.command "constdefs" "old-style constant definition" K.thy_decl
    (structs -- Scan.repeat1 old_constdef >> (Toplevel.theory o add_constdefs'));







(* *** Nonterminals *** *)

fun add_nonterminals' ntl thy =
  let val thynm = Aux.name_of_thy thy
      fun save_nt nt = 
         let val _ = Syn_mng.add_nonterm (thynm, Binding.name_of nt)
         in 
          nt
         end
  in
   Sign.add_nonterminals (map save_nt ntl) thy
  end

val _ =
  OuterSyntax.command "nonterminals" "declare types treated as grammar nonterminal symbols"
    K.thy_decl (Scan.repeat1 P.binding >> (Toplevel.theory o add_nonterminals'));



(* *** Syntax *** *)

fun add_modesyntax' ((s, b), sl) thy =
  let val thynm = Aux.name_of_thy thy
      fun save_syn (synnm, typestr, mxfx) = 
         let val typ = ProofContext.read_typ_syntax (Context.init_proof thy) typestr
         in
           Syn_mng.add_syn_decl (thynm, (s, b, synnm, typ, mxfx))
         end
      val _ = map save_syn sl
  in
   (uncurry Sign.add_modesyntax) ((s, b), sl) thy
  end

val _ =
  OuterSyntax.command "syntax" "declare syntactic constants" K.thy_decl
    (opt_mode -- Scan.repeat1 P.const >> (Toplevel.theory o add_modesyntax')) 
                                           

(* *** Translations *** *)

fun add_trrules' trl thy =
 let val thyname = Aux.name_of_thy thy
     fun save_trrule tr =
         let val _ = Syn_mng.add_trrule (thyname, tr)
         in
          tr
         end
 in
  Sign.add_trrules (map save_trrule trl) thy
 end  

val trans_pat =
  Scan.optional (P.$$$ "(" |-- P.!!! (P.xname --| P.$$$ ")")) "logic" -- P.string;

fun trans_arrow toks =
  ((P.$$$ "\\<rightharpoonup>" || P.$$$ "=>") >> K Syntax.ParseRule ||
    (P.$$$ "\\<leftharpoondown>" || P.$$$ "<=") >> K Syntax.PrintRule ||
    (P.$$$ "\\<rightleftharpoons>" || P.$$$ "==") >> K Syntax.ParsePrintRule) toks;

val trans_line =
  trans_pat -- P.!!! (trans_arrow -- trans_pat)
    >> (fn (left, (arr, right)) => arr (left, right));

val _ =
  OuterSyntax.command "translations" "declare syntax translation rules" K.thy_decl
    (Scan.repeat1 trans_line >> (Toplevel.theory o add_trrules'));





(* *** Translation functions *** *)

fun parse_ast_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_parse_ast_trans (thyname, (b, s))
 in
  IsarCmd.parse_ast_translation (b, s) thy
 end

fun parse_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_parse_trans (thyname, (b, s))
 in
  IsarCmd.parse_translation (b, s) thy
 end

fun print_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_print_trans (thyname, (b, s))
 in
  IsarCmd.print_translation (b, s) thy
 end

fun print_ast_translation' (b, s) thy =
 let val thyname = Aux.name_of_thy thy
     val _ = Syn_mng.add_print_ast_trans (thyname, (b, s))
 in
  IsarCmd.print_ast_translation (b, s) thy
 end



val trfun = P.opt_keyword "advanced" -- P.position P.text;

val _ =
  OuterSyntax.command "parse_ast_translation" "install parse ast translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o parse_ast_translation'));

val _ =
  OuterSyntax.command "parse_translation" "install parse translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o parse_translation'));

val _ =
  OuterSyntax.command "print_translation" "install print translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o print_translation'));



val _ =
  OuterSyntax.command "print_ast_translation" "install print ast translation functions"
    (K.tag_ml K.thy_decl)
    (trfun >> (Toplevel.theory o print_ast_translation'));



(* *** Theory *** *)

fun theory' (s, sl, l) t =
  let val t' = IsarCmd.init_theory (s, sl, l) t
      val _ = Syn_mng.upd_rem_tabs s
  in
   t'
  end

val _ =
  OuterSyntax.command "theory" "begin theory" (K.tag_theory K.thy_begin)
    (ThyHeader.args >> (Toplevel.print oo theory'));


end