(* 
 * Extension of some Isabelle/HOL specific parsers
 * $Id: parser_ext_hol.sml 1162 2009-04-21 09:41:10Z maxim $
 *)


structure Parser_Ext_HOL =
struct


structure P = OuterParse and K = OuterKeyword;

(* ********************************* *)
(* ***    HOL-specific syntax    *** *)
(* ********************************* *)


(* *** definition  *** *)


fun add_definition (opt, x) lthy =
  (case opt of
    NONE => Specification.definition_cmd (opt, x) lthy
  | SOME d => (let val _ = Parser_Ext.save_mixfix ((ProofContext.theory_of o LocalTheory.target_of) lthy) d
               in
                Specification.definition_cmd (opt, x) lthy
               end))






val constdecl =
  P.binding --
    (P.where_ >> K (NONE, NoSyn) ||
      P.$$$ "::" |-- P.!!! ((P.typ >> SOME) -- P.opt_mixfix' --| P.where_) ||
      Scan.ahead (P.$$$ "(") |-- P.!!! (P.mixfix' --| P.where_ >> pair NONE))
  >> P.triple2;

val constdef = Scan.option constdecl -- (SpecParse.opt_thm_name ":" -- P.prop);

val _ =
  OuterSyntax.local_theory "definition" "constant definition" K.thy_decl
    (constdef >> (fn args => #2 o add_definition args));





(* *************** *)
(* *** Records *** *)
(* *************** *)

(* Orig.
val record_decl =
  P.type_args -- P.name --
    (P.$$$ "=" |-- Scan.option (P.typ --| P.$$$ "+") -- Scan.repeat1 P.const);

val _ =
  OuterSyntax.command "record" "define extensible record" K.thy_decl
    (record_decl >> (fn (x, (y, z)) => Toplevel.theory (add_record false x y z)));

*)

fun add_record' a b c thy = (let val thynm = Aux.name_of_thy thy
                                 fun mk_nm s = (thynm ^ "." ^ (snd a) ^ "." ^ s)
                                 fun add_mxfxs [] = ()
                                   | add_mxfxs ((nm, tp, mxfx)::xs) =
                                      (let val fnm = mk_nm nm
                                           val _ = (case mxfx of
                                                     NoSyn => ()
                                                    | _ => Syn_mng.add_mixfix (fnm, mxfx))
                                       in
                                        add_mxfxs xs
                                       end)
                                 val _ = add_mxfxs c
                             in
                              RecordPackage.add_record false a b c thy
                             end) 

val record_decl =
  P.type_args -- P.name --
    (P.$$$ "=" |-- Scan.option (P.typ --| P.$$$ "+") -- Scan.repeat1 P.const);

val _ =
  OuterSyntax.command "record" "define extensible record" K.thy_decl
    (record_decl >> (fn (x, (y, z)) => Toplevel.theory (add_record' x y z)));



(* *** Datatypes *** *)

(* <thynm>.<typenm>.<consnm> *)

(* Orig.

val datatype_decl =
  Scan.option (P.$$$ "(" |-- P.name --| P.$$$ ")") -- P.type_args -- P.binding -- P.opt_infix --
    (P.$$$ "=" |-- P.enum1 "|" (P.binding -- Scan.repeat P.typ -- P.opt_mixfix));

fun mk_datatype args =
  let
    val names = map
      (fn ((((NONE, _), t), _), _) => Binding.name_of t | ((((SOME t, _), _), _), _) => t) args;
    val specs = map (fn ((((_, vs), t), mx), cons) =>
      (vs, t, mx, map (fn ((x, y), z) => (x, y, z)) cons)) args;
  in snd o add_datatype_cmd false names specs end;

val _ =
  OuterSyntax.command "datatype" "define inductive datatypes" K.thy_decl
    (P.and_list1 datatype_decl >> (Toplevel.theory o mk_datatype));


*)

val datatype_decl =
  Scan.option (P.$$$ "(" |-- P.name --| P.$$$ ")") -- P.type_args -- P.binding -- P.opt_infix --
    (P.$$$ "=" |-- P.enum1 "|" (P.binding -- Scan.repeat P.typ -- P.opt_mixfix));


 
fun mk_datatype' args thy =
  let val names = map (fn ((((NONE, _), t), _), _) => Binding.name_of t | ((((SOME t, _), _), _), _) => t) args;
      val specs = map (fn ((((_, vs), t), mx), cons) =>
                   (vs, t, mx, map (fn ((x, y), z) => (x, y, z)) cons)) args;
      val result = snd(DatatypePackage.add_datatype_cmd false names specs thy);

      val thynm = Aux.name_of_thy thy; 
      fun add_mixfix [] = () 
        | add_mixfix ((vs, bind_t, mx, cons)::xs) =
          (let val _ = (case mx of
                         NoSyn => ()
                       | _ => Syn_mng.add_type_mixfix (Sign.full_name thy bind_t, mx)) (*(thynm ^ "." ^ t, mx))*)
          
               fun add_cons_mixfix [] = ()
                 | add_cons_mixfix ((bind_nm, b, mx)::ys) =
                   (let val _ = (case mx of
                                   NoSyn => ()
                                 | _ => Syn_mng.add_mixfix (thynm ^ "." ^ (Binding.name_of bind_t) ^ "." ^ 
                                                                          (Binding.name_of bind_nm), 
                                                            mx))
                    in
                     add_cons_mixfix ys
                    end)
                val _ = add_cons_mixfix cons;
           in 
             add_mixfix xs
           end);
      val _ = add_mixfix specs;
  in 
   result 
  end;


val _ =
  OuterSyntax.command "datatype" "define inductive datatypes" K.thy_decl
    (P.and_list1 datatype_decl >> (Toplevel.theory o mk_datatype'));


(* *** Function Package *** *)

fun add_fundef_cmd' fixes statements config lthy =
(let val thy = (ProofContext.theory_of o LocalTheory.target_of) lthy
     val _ = map (Parser_Ext.save_mixfix thy) fixes
 in
  FundefPackage.add_fundef_cmd fixes statements config lthy 
 end)


(* Orig. 

val _ =
  OuterSyntax.local_theory_to_proof "function" "define general recursive functions" K.thy_goal
  (fundef_parser default_config
     >> (fn ((config, fixes), statements) => add_fundef_cmd fixes statements config));

*)


val _ =
  OuterSyntax.local_theory_to_proof "function" "define general recursive functions" K.thy_goal
  (FundefCommon.fundef_parser FundefCommon.default_config
     >> (fn ((config, fixes), statements) => add_fundef_cmd' fixes statements config))



 (* Fun - command *)

(* Orig.

fun gen_fun add config fixes statements int lthy =
  let val group = serial_string () in
    lthy
      |> LocalTheory.set_group group
      |> add fixes statements config
      |> by_pat_completeness_auto int
      |> LocalTheory.restore
      |> LocalTheory.set_group group
      |> termination_by (FundefCommon.get_termination_prover lthy) int
  end;

val add_fun = gen_fun FundefPackage.add_fundef
val add_fun_cmd = gen_fun FundefPackage.add_fundef_cmd



local structure P = OuterParse and K = OuterKeyword in

val _ =
  OuterSyntax.local_theory' "fun" "define general recursive functions (short version)" K.thy_decl
  (fundef_parser fun_config
     >> (fn ((config, fixes), statements) => add_fun_cmd config fixes statements));


*)

fun pat_completeness ctxt = SIMPLE_METHOD' (FundefDatatype.pat_completeness_tac ctxt)

val by_pat_completeness_auto =
    Proof.global_future_terminal_proof
      (Method.Basic (pat_completeness, Position.none),
       SOME (Method.Source_i (Args.src (("HOL.auto", []), Position.none))))

fun termination_by method int =
    FundefPackage.termination_proof NONE
    #> Proof.global_future_terminal_proof
      (Method.Basic (method, Position.none), NONE) int



fun gen_fun add config fixes statements int lthy =
  let val group = serial_string () in
    lthy
      |> LocalTheory.set_group group
      |> add fixes statements config 
      |> by_pat_completeness_auto int
      |> LocalTheory.restore
      |> LocalTheory.set_group group
      |> termination_by (FundefCommon.get_termination_prover lthy) int
  end;


fun add_fun_cmd' config fixes statements int lthy =
(let val thy = (ProofContext.theory_of o LocalTheory.target_of) lthy
     val _ = map (Parser_Ext.save_mixfix thy) fixes
 in
  gen_fun FundefPackage.add_fundef_cmd config fixes statements int lthy
 end)


val fun_config = FundefCommon.FundefConfig 
                  { sequential=true, default="%x. arbitrary" (* FIXME dynamic scoping *), 
                    domintros=false, tailrec=false }


val _ =
  OuterSyntax.local_theory' "fun" "define general recursive functions (short version)" K.thy_decl
  (FundefCommon.fundef_parser fun_config
     >> (fn ((config, fixes), statements) => add_fun_cmd' config fixes statements));


end